require "defines"

local floor = math.floor
local min = math.min

-- Constants
local FRAME_NAME = "dark_game_clock"
local VALID_LOCATIONS = {
  left = true,
  top = true,
  center = true
 }
local UPDATE_INTERVAL = 6 -- minimal update interval
-- proper types for settings
local SETTINGS = {
  enabled = "boolean",
  show_clock = "boolean",
  show_run_clock = "boolean",
  style = "string",
  minute_rounding = "number",
  location = "string",
  show_minutes = "boolean",
  show_seconds_in_runtime = "boolean",
  frame_leftpadding = "number"
}

local DEFAULT_SETTINGS = {
  enabled = true,
  show_clock = true,
  show_run_clock = true,
  style = "24h",
  minute_rounding = 15,
  location = "left",
  show_minutes = true,
  show_seconds_in_runtime = false,
  frame_leftpadding = 20
}

-- gui stuff
local function remove_gui(playerindex)
  local player = game.players[playerindex]
  --if not player then error("Interal Error: Invalid player index in remove_gui") end
  local nickname = player.name
  if global[nickname].frame and global[nickname].frame.valid then
    global[nickname].frame.destroy()
  else
    -- frame not valid in global after load (or isn't properly saved, didn't actually test which)....
    for loc, _ in pairs(VALID_LOCATIONS) do
      if player.gui[loc][FRAME_NAME] then player.gui[loc][FRAME_NAME].destroy() end
    end
  end
end

local function init_gui(playerindex)
  remove_gui(playerindex)
  
  local player = game.players[playerindex]
  --if not player then error("Internal Error: invalid player index in init_gui") end
  local playerSettings = global[player.name]
  if not playerSettings.enabled then return end
  
  
  frame = player.gui[playerSettings.location].add{type="frame", name=FRAME_NAME, caption = '', style='naked_frame_style', direction='vertical'}
  frame.style.left_padding = playerSettings.frame_leftpadding
  
  if playerSettings.show_clock then
    frame.add{type="label", name="game_clock", caption=""}
    local style_game = frame.game_clock.style
    style_game.font = "default-frame"
  end
  
  if playerSettings.show_run_clock then
    frame.add{type="label", name="run_clock", caption=""}
  end
  
  -- seems a little silly to store the actual reference in what's called
  -- playerSettings but... it's the only thing that isn't a setting so...
  -- it'd seem even sillier right now to add an entire 'settings' table
  playerSettings.frame = frame
  
end

local function do_update(playerindex)
  local player = game.players[playerindex]
  --if not player then error("Internal Error: invalid player index in do_update") end
  local nickname = player.name
  local playerSettings = global[nickname]
  if not (playerSettings.frame and playerSettings.frame.valid) then return end
  
  if playerSettings.show_clock and playerSettings.frame.game_clock then
    local time_normal = game.daytime*24
    -- flip time around, 0.5 is midnight ugh
    if time_normal > 12 then time_normal = time_normal - 12 else time_normal = time_normal + 12 end

    local ampm = ""
    local minute = ""
    local hour = floor(time_normal)

    if playerSettings.show_minutes then
      local raw_mins = floor((time_normal % 1)*60)
      minute = string.format(":%02d", raw_mins - raw_mins % playerSettings.minute_rounding )
    end

    if playerSettings.style == "12h" then 
      ampm = (hour >= 12 and " PM") or " AM"
      hour = (hour == 0 and 12) or (hour > 12 and (hour - 12)) or hour
    end

    playerSettings.frame.game_clock.caption = string.format("%02d%s%s", hour, minute, ampm)
  end
  
  if playerSettings.show_run_clock and playerSettings.frame.run_clock then
    local run_time_s = floor(game.tick/60)
    local run_time_minutes = floor(run_time_s/60)
    local run_time_hours = floor(run_time_minutes/60)
    local run_time_days = floor(run_time_hours/24)
    local days = ""
    if run_time_days > 0 then days = string.format("%d days, ", run_time_days) end
    local seconds = ""
    if playerSettings.show_seconds_in_runtime then
      seconds = string.format(":%02d", run_time_s % 60)
    end
    playerSettings.frame.run_clock.caption = string.format("%s%02d:%02d%s", days, run_time_hours % 24, run_time_minutes % 60, seconds)
  end

end 

local function set_defaults(playerindex)
  local player = game.players[playerindex]
  local nickname = player.name
  global[nickname] = global[nickname] or DEFAULT_SETTINGS
end

-- events
script.on_event(defines.events.on_player_created, function(event)
  set_defaults(event.player_index)
  init_gui(event.player_index)
  do_update(event.player_index)
end)

script.on_load(function()
  -- restore gui state on load or right after save
  for playerindex, player in ipairs(game.players) do
    set_defaults(playerindex)
    init_gui(playerindex)
    do_update(playerindex)
  end
end)



--[[
  to manually remove use this (copy/paste) at the console
  
  In singleplayer use:
  /c for _, location in ipairs({"center", "top", "left"}) do if game.player.gui[location].dark_game_clock then game.player.gui[location].dark_game_clock.destroy() end end
  
  In multiplayer use this for all players:
  /c for index, player in pairs(game.players) do if player.name == "my_nick" then for _, location in ipairs({"center", "top", "left"}) do if player.gui[location].dark_game_clock then player.gui[location].dark_game_clock.destroy() end end end end
  
  In multiplayer to only affect yourself (for whatever reason) use:
  first change my_nick to whatever your multiplayer nickname is however!
  
  /c for index, player in pairs(game.players) do if player.name == "my_nick" then for _, location in ipairs({"center", "top", "left"}) do if player.gui[location].dark_game_clock then player.gui[location].dark_game_clock.destroy() end end end end
  
--]]
--[[ caused issues in multiplayer
game.onsave(function()
  -- alow mod to be removed along with gui at any time
  for playerindex, player in ipairs(game.players) do
    remove_gui(playerindex)
  end
end)
--]]

script.on_event(defines.events.on_tick, function(event)
  for playerindex, player in ipairs(game.players) do
    local settings = global[player.name]
    if not (settings and (not settings.enabled or (settings.frame and settings.frame.valid))) then
      set_defaults(playerindex)
      init_gui(playerindex)
    end
    if (event.tick % (UPDATE_INTERVAL * global[player.name].minute_rounding) == 0) then
      do_update(playerindex)
    end
  end
end)

-- helper functions
local function printToAll(msg)
  for _, player in ipairs(game.players) do
    player.print(msg)
  end
end

local function indexFromNickname(nickname)
  if not nickname or not type(nickname) == "string" then
    printNicknameError(nickname)
    return nil
  end
  for index, player in ipairs(game.players) do
    -- perhaps this should be player.name:lower() == nickname:lower() ?
    if player.name == nickname then
      return index
    end
  end
  -- explicit return nil, nickname not found!
  return nil
end

function printNicknameError(nickname)
  if type(nickname) == "table" then
    nickname = serpent.block(nickname)
  else
    nickname = tostring(nickname)
  end
  printToAll({"unknown-nickname", nickname})
  printToAll({"check-case"})
end

-- remote interface
remote.add_interface("TimeDisplay", {
  -- remote.call("TimeDisplay", "SetSetting", "setting_name", value, "nickname")
  --[[
      -- controls mod state                             [default: true]
  /c remote.call("TimeDisplay", "SetSetting", "enabled", true/false, "nickname")
      -- controls display of day time clock             [default: true]
  /c remote.call("TimeDisplay", "SetSetting", "show_clock", true/false, "nickname")
      -- controls display of run time clock             [default: true]
  /c remote.call("TimeDisplay", "SetSetting", "show_run_clock", true/false, "nickname")
      -- style for day time clock                       [default: 24h]
  /c remote.call("TimeDisplay", "SetSetting", "style", "12h/24h", "nickname")
      -- display location                               [default: left]
  /c remote.call("TimeDisplay", "SetSetting", "location", "left/top/center", "nickname")
      -- rounds minutes in daytime clock to this value  [default: 15]
  /c remote.call("TimeDisplay", "SetSetting", "minute_rounding", 1-60, "nickname")
      -- show/hide minutes in day time clock            [default: true]
  /c remote.call("TimeDisplay", "SetSetting", "show_minutes", true/false, "nickname")
      -- should run time clock show seconds             [default: false]
  /c remote.call("TimeDisplay", "SetSetting", "show_seconds_in_runtime", true/false, "nickname")
      -- shifts display horizontally, set to 1450 on 1080p monitor to move display near minimap  [default: 20]
  /c remote.call("TimeDisplay", "SetSetting", "frame_leftpadding", 0-9999, "nickname")
  ]]--
  SetSetting = function(setting, value, nickname)
    local playerindex = indexFromNickname(nickname)
    if not playerindex then
      printNicknameError(nickname)
      return
    end
    local player = game.players[playerindex]
    if not SETTINGS[setting] then
      player.print({"unknown-setting", setting})
      return
    end
    -- check type
    if type(value) ~= SETTINGS[setting] then
      player.print({"value-missmatch", SETTINGS[setting], type(value)})
      return
    end
    -- check values
    if setting == "style" and (value ~= "12h" and value ~= "24h") then
      player.print({"unexpected-value", "12h or 24h", value})
      return
    end    
    if setting == "location" and not VALID_LOCATIONS[value] then
      player.print({"unexpected-value", "left, top or center", value})
      return
    end
    if setting == "minute_rounding" and (value > 60 or value < 1) then
      player.print({"unexpected-value", "1-60", value})
      return
    end
    
    -- appply
    if not global[nickname] then
      set_defaults(playerindex)
    end
    global[nickname][setting] = value
    init_gui(playerindex)
    do_update(playerindex)
  end
})
